#!/bin/python3
#%%
import datetime
import os
import re
import time
import zlib

#%%

replace_slash = lambda p: p.replace('\\', '/')


def all_in_dir(dir_path, ls = []):
    ls_dir = os.listdir(dir_path)
    if ls:
        if type(ls) == list:
            if len(set(ls) - set(ls_dir)) == 0:
                r = True
            else:
                r = False
        elif type(ls) == str:
            pattern = ls
            r = True
            for p in ls_dir:
                if not re.findall(pattern, p):
                    r = False
                    break
        else:
            print('ERROR: all_in_dir() type error')
    else:
        r = True
    return r


def any_in_dir(dir_path, ls = []):
    ls_dir = os.listdir(dir_path)
    if ls:
        if type(ls) == list:
            if len(set(ls) - set(ls_dir)) < len(ls):
                r = True
            else:
                r = False
        elif type(ls) == str:
            pattern = ls
            r = False
            for p in ls_dir:
                if re.findall(pattern, p):
                    r = True
                    break
        else:
            print('ERROR: any_in_dir(): type error')
    else:
        r = True
    return r


def ls_in_dir(dir_path, all_ls = [], any_ls = []):
    r1 = all_in_dir(dir_path, all_ls)
    r2 = any_in_dir(dir_path, any_ls)
    r = r1 & r2
    return r


def ls_files_r(path = './', slash = False, pattern = '', show = False):
    if os.path.exists(path):
        r = []
        for k1, _, k2 in os.walk(path):
            for k22 in k2:
                if re.findall(pattern, k22, re.IGNORECASE):
                    p = os.path.join(k1, k22)
                    r.append(p)
        r = sorted(r)
        if slash:
            r = list(map(replace_slash, r))
        if show:
            [print(f"path = '{k}'") for k in r]
    else:
        print(f'ERROR: {path} does not exist.')
    return r


def ls_dirs_r(path = './', slash = False, all_ls = [], any_ls = [], pattern = '', show = False):
    if os.path.exists(path):
        r = [path] if re.findall(pattern, path) else []
        for k1, k2, _ in os.walk(path):
            # r.append(k1)
            for k22 in k2:
                if re.findall(pattern, k22, re.IGNORECASE):
                    p = os.path.join(k1, k22)
                    r.append(p)
        r = sorted(r)
        if slash:
            r = list(map(replace_slash, r))
        r = [p for p in r if ls_in_dir(p, all_ls, any_ls)]
        if show:
            [print(f"path = '{k}'") for k in r]
    else:
        print(f'ERROR: {path} does not exist.')
    return r


def ls_files(path, slash = False, pattern = '', show = False):
    ps = sorted(os.listdir(path))
    paths = [os.path.join(path, k) for k in ps if re.findall(pattern, k)]
    dirs = []
    for p in paths:
        if os.path.isfile(p):
            dirs.append(p)
    if slash:
        dirs = [replace_slash(p) for p in dirs]
    if show:
        [print(f"path = '{k}'") for k in dirs]
    return dirs


def ls_dirs(path, slash = False, all_ls = [], any_ls = [], pattern = '', show = False):
    ps = sorted(os.listdir(path))
    paths = [os.path.join(path, k) for k in ps if re.findall(pattern, k)]
    dirs = []
    for p in paths:
        if os.path.isdir(p):
            dirs.append(p)
    if slash:
        dirs = [replace_slash(p) for p in dirs]
    dirs = [p for p in dirs if ls_in_dir(p, all_ls, any_ls)]
    if show:
        [print(f"path = '{k}'") for k in dirs]
    return dirs


def find_latest_file(path: str, pattern: str):
    latest_timestamp = 0
    latest_file = None
    for file in os.listdir(path):
        if re.findall(pattern, file):
            full_path = os.path.join(path, file)
            timestamp = os.path.getmtime(full_path)
            if timestamp > latest_timestamp:
                latest_timestamp = timestamp
                latest_file = full_path
    print(f'latest_file = {latest_file}')
    return latest_file


#%%

def add(a, b):
    c = round(a + b, 6)
    a, b = round(a, 6), round(b, 6)
    string = f'{a} + {b} = {c}'
    return string


def subtract(a, b):
    c = round(a - b, 6)
    a, b = round(a, 6), round(b, 6)
    string = f'{a} - {b} = {c}'
    return string


def multiply(a, b):
    c = round(a * b, 6)
    a, b = round(a, 6), round(b, 6)
    string = f'{a} * {b} = {c}'
    return string

def divide(a, b, units = None):
    if b == 0:
        c = 'inf'
    else:
        c = round(a / b, 6)
    a, b = round(a, 6), round(b, 6)
    if units is None:
        string = f'{a} / {b} = {c}'
    else:
        string = f'{a} {units[0]} / {b} {units[1]} = {c} {units[0]}/{units[1]}'
    return string


def subtract_with_divide(a, b):
    c = round(a - b, 6)
    if b == 0:
        d = 'inf'
    else:
        d = round(b / a, 6)
    string = f'{a} - {b} = {c}, 1 - {b}/{a}({d}) = {1-d}'
    return string


def frac(a, b, units = ['', '']):
    string = f'{a}{units[0]}/{b}{units[1]} = {a / b}'
    return string


def check_match(pattern: str, string: str, validate = False):
    ret = False
    if len(pattern) > 0 and len(string) > 0:
        matches = re.findall(pattern, string)
        if validate and len(matches) != 1:
            raise ValueError(f'matches = {matches}')
        if len(matches) > 0:
            ret = True
    elif len(pattern) == 0:
        ret  =True
    return ret


def find_string(pattern: str, string: str, validate = True):
    """
    Finds the first occurrence of a pattern in a given string.

    Parameters:
        pattern (str): The regular expression pattern to search for.
        string (str): The string to search in.
        validate (bool): If True, raises an error if the number of matches is not 1. Defaults to True.

    Returns:
        str: The first match of the pattern in the string, or an empty string if no match is found.
    """
    matches = re.findall(pattern, string)
    if validate and len(matches) != 1:
        raise ValueError(f'matches = {matches}')
    if len(matches) > 0:
        result = matches[0]
    else:
        result = ''
    return result


def zlib_compress(data: bytes, **kwargs):
    compressed_data = zlib.compress(data, **kwargs)
    print(f'compressed_ratio = {frac(len(compressed_data), len(data))}')
    return compressed_data


def zlib_decompress(data: bytes, **kwargs):
    decompress_data = zlib.decompress(data, **kwargs)
    print(f'compressed_ratio = {frac(len(data), len(decompress_data))}')
    return decompress_data


def read_binary(path: str):
    with open(path, 'rb') as f:
        data = f.read()
    return data




def find_matching_paths(
        root_path: str, 
        pattern: str = ".*", 
        search_type: str = "both", 
        expected_count: int = None, 
        find_first: bool = False, 
        raise_error: bool = True, 
        print_paths: bool = False,
        max_depth: int = None,
        ):
    
    """
    Find files or directories in a path that partially match a regular expression.

    Parameters:
    - root_path (str): The root path to search.
    - pattern (str): Regular expression pattern for file or directory names.
    - search_type (str): Specify the type to search for; options are "file", "dir", or "both".
    - expected_count (int): Expected number of matching results. If it does not match, an error is raised if raise_error is True.
    - find_first (bool): If True, stops searching after finding the first match and returns the path.
    - raise_error (bool): If True, raises an error if the number of matches does not equal expected_count.
    - print_paths (bool): If True, prints the matching paths.
    - max_depth (int): Maximum depth of subdirectories to search.

    Returns:
    - list: List of paths for files or directories that match the criteria.
    """

    if not os.path.isdir(root_path):
        raise ValueError("Root path must be a directory.")
    
    if root_path[-1] in ["\\", "/"]:
        root_path = root_path[:-1]

    # Ensure search_type is not "both" when find_first is True
    if find_first and search_type == "both":
        raise ValueError("When find_first is True, search_type cannot be 'both'.")

    matching_paths = []
    regex = re.compile(pattern)

    for dirpath, dirnames, filenames in os.walk(root_path):

        if max_depth is not None:
            depth = dirpath[len(root_path):].count(os.sep)
            if depth >= max_depth:
                continue

        # Search matching directories if specified in search_type
        if search_type in ["dir", "both"]:
            for dirname in dirnames:
                if regex.search(dirname):  # Partial match
                    matching_paths.append(os.path.join(dirpath, dirname))
                    if find_first:
                        break
            if find_first and matching_paths:
                break

        # Search matching files if specified in search_type
        if search_type in ["file", "both"]:
            for filename in filenames:
                if regex.search(filename):  # Partial match
                    matching_paths.append(os.path.join(dirpath, filename))
                    if find_first:
                        break
            if find_first and matching_paths:
                break

    # Print matching paths if print_paths is True
    if print_paths:
        print("Matching paths found:")
        for path in matching_paths:
            print(path)

    # Check if the number of results matches the expected count
    if expected_count is not None and len(matching_paths) != expected_count:
        error_message = f"Number of matching paths is {len(matching_paths)}, but expected {expected_count}."
        if raise_error:
            raise ValueError(error_message)
        else:
            print("Warning:", error_message)

    # Return matching paths
    return matching_paths


#%%

import json

json_dump = lambda d, path: json.dump(
    d, 
    open(path, 'w', encoding='utf-8', newline='\n'), 
    indent=4, 
    ensure_ascii=False
    )
json_dumps = lambda d, indent=4: json.dumps(
    d, 
    indent=indent, 
    ensure_ascii=False
    )
json_load = lambda path: json.load(
    open(path, 'r', encoding='utf-8')
    )
json_loads = lambda s: json.loads(
    s
    )


#%%

def read_file_with_fallback(file_path):
    encodings = ['utf-8', 'gbk']
    content = None
    for encoding in encodings:
        try:
            with open(file_path, 'r', encoding=encoding) as f:
                content = f.read()
            print(f"Successfully decoded using {encoding}")
            break
        except UnicodeDecodeError:
            print(f"Failed to decode using {encoding}")
    return content

#%%
import re
from typing import Union, List

def replace_in_text_files(
    file_paths: Union[str, List[str]],
    pattern: str,
    replacement: str,
    output_files: Union[None, str, List[str]] = None,
    validate_single_match: bool = False
):
    """
    Replace text in files that match a given regular expression. Supports single or multiple files.
    
    Parameters:
        file_paths (Union[str, List[str]]): Path to a single file or a list of file paths.
        pattern (str): Regular expression to match text in the file(s).
        replacement (str): String to replace the matched text.
        output_files (Union[None, str, List[str]]): Path(s) to save the modified file(s).
            - If None, overwrite the original file(s).
            - If a single string, save all files to the same path.
            - If a list, it should correspond to file_paths one-to-one.
        validate_single_match (bool): If True, raise an error if the number of matches is not exactly one.
    """
    # Convert a single string file path to a list for unified processing
    if isinstance(file_paths, str):
        file_paths = [file_paths]
    
    if output_files is None:
        output_files = [None] * len(file_paths)
    elif isinstance(output_files, str):
        output_files = [output_files] * len(file_paths)

    if len(file_paths) != len(output_files):
        raise ValueError("The number of input files does not match the number of output files.")

    for file_path, output_file in zip(file_paths, output_files):
        try:
            # Read file content
            with open(file_path, 'r', encoding='utf-8') as file:
                content = file.read()
            
            # Find all matches for validation
            matches = re.findall(pattern, content)
            if validate_single_match:
                if len(matches) != 1:
                    raise ValueError(
                        f"File {file_path} contains {len(matches)} match(es) for the pattern, "
                        f"but exactly one match is required."
                    )
            
            # Replace matched content
            new_content = re.sub(pattern, replacement, content)

            # Determine the save path
            save_path = output_file if output_file else file_path

            # Save the modified content
            with open(save_path, 'w', encoding='utf-8') as file:
                file.write(new_content)

            print(f"File successfully saved to {save_path}.")
        except Exception as e:
            print(f"An error occurred while processing the file {file_path}: {e}")



#%%

def human_readable_size(file_size: int):
    if file_size < 0:
        raise ValueError(f'file_size = {file_size} < 0')
    size = file_size
    unit_index = 0
    units = ['B', 'KiB', 'MiB', 'GiB', 'TiB']
    while size >= 1024 and unit_index < len(units) - 1:
        size /= 1024.0
        unit_index += 1
    ret = f'{size:.2f} {units[unit_index]}'
    return ret


def get_timestamp(dt: datetime.datetime = None):
    time_format = "%Y%m%d%a%H%M%S"
    if dt is None:
        result = time.strftime(time_format, time.localtime())
    elif isinstance(dt, datetime.datetime):
        result = dt.strftime(time_format)
    return result

#%%

def count_down(n: int):
    if type(n) == list:
        s = sorted(set(n))
        n = (s[1] - s[0]) * random() + s[0] + 1
    n = int(n)
    for k in range(n):
        print(f'{n - k - 1}s  \r', end = '')
        time.sleep(1)
    return None

#%%
import uuid

get_uuid_hex = lambda : uuid.uuid4().hex
